@@ -40,7 +40,8 @@
 #include "../../_psutil_common.h"
 
 #define PSUTIL_KPT2DOUBLE(t) (t ## _sec + t ## _usec / 1000000.0)
-#define PSUTIL_TV2DOUBLE(t) ((t).tv_sec + (t).tv_usec / 1000000.0)
+// #define PSUTIL_TV2DOUBLE(t) ((t).tv_sec + (t).tv_usec / 1000000.0)
+
 // a signaler for connections without an actual status
 int PSUTIL_CONN_NONE = 128;
 
@@ -239,7 +240,11 @@ psutil_get_cmdline(long pid) {
         goto error;
 
     for (p = argv; *p != NULL; p++) {
+#if PY_MAJOR_VERSION >= 3
+        py_arg = PyUnicode_DecodeFSDefault(*p);
+#else
         py_arg = Py_BuildValue("s", *p);
+#endif
         if (!py_arg)
             goto error;
         if (PyList_Append(py_retlist, py_arg))
@@ -466,7 +471,11 @@ psutil_proc_cwd(PyObject *self, PyObject *args) {
         PyErr_SetFromErrno(PyExc_OSError);
         return NULL;
     }
+#if PY_MAJOR_VERSION >= 3
+    return PyUnicode_DecodeFSDefault(path);
+#else
     return Py_BuildValue("s", path);
+#endif
 }
 
 
@@ -764,15 +773,12 @@ psutil_disk_io_counters(PyObject *self, PyObject *args) {
 
     for (i = 0; i < dk_ndrive; i++) {
         py_disk_info = Py_BuildValue(
-            "(KKKKLL)",
-            stats[i].ds_rxfer,
-            stats[i].ds_wxfer,
-            stats[i].ds_rbytes,
-            stats[i].ds_wbytes,
-            // assume half read - half writes.
-            // TODO: why?
-            (long long) PSUTIL_TV2DOUBLE(stats[i].ds_time) / 2,
-            (long long) PSUTIL_TV2DOUBLE(stats[i].ds_time) / 2);
+            "(KKKK)",
+            stats[i].ds_rxfer,  // num reads
+            stats[i].ds_wxfer,  // num writes
+            stats[i].ds_rbytes,  // read bytes
+            stats[i].ds_wbytes  // write bytes
+        );
         if (!py_disk_info)
             goto error;
         if (PyDict_SetItemString(py_retdict, stats[i].ds_name, py_disk_info))
@@ -791,3 +797,27 @@ error:
     return NULL;
 }
 
+
+PyObject *
+psutil_cpu_stats(PyObject *self, PyObject *args) {
+    size_t size;
+    struct uvmexp uv;
+    int uvmexp_mib[] = {CTL_VM, VM_UVMEXP};
+
+    size = sizeof(uv);
+    if (sysctl(uvmexp_mib, 2, &uv, &size, NULL, 0) < 0) {
+        PyErr_SetFromErrno(PyExc_OSError);
+        return NULL;
+    }
+
+    return Py_BuildValue(
+        "IIIIIII",
+        uv.swtch,  // ctx switches
+        uv.intrs,  // interrupts - XXX always 0, will be determined via /proc
+        uv.softs,  // soft interrupts
+        uv.syscalls,  // syscalls - XXX always 0
+        uv.traps,  // traps
+        uv.faults,  // faults
+        uv.forks  // forks
+    );
+}
